from __future__ import annotations

from timmx.export.base import ExportBackend
from timmx.export.coreml_backend import CoreMLBackend
from timmx.export.executorch_backend import ExecuTorchBackend
from timmx.export.litert_backend import LiteRTBackend
from timmx.export.ncnn_backend import NcnnBackend
from timmx.export.onnx_backend import OnnxBackend
from timmx.export.tensorrt_backend import TensorRTBackend
from timmx.export.torch_export_backend import TorchExportBackend
from timmx.export.torchscript_backend import TorchScriptBackend


class BackendRegistry:
    """In-repo backend registry used by the CLI."""

    def __init__(self) -> None:
        self._backends: dict[str, ExportBackend] = {}

    def register(self, backend: ExportBackend) -> None:
        if backend.name in self._backends:
            raise ValueError(f"Backend {backend.name!r} is already registered.")
        self._backends[backend.name] = backend

    def get(self, name: str) -> ExportBackend:
        try:
            return self._backends[name]
        except KeyError as exc:
            raise KeyError(f"Unknown backend {name!r}.") from exc

    def items(self) -> list[tuple[str, ExportBackend]]:
        return sorted(self._backends.items(), key=lambda item: item[0])

    def names(self) -> list[str]:
        return [name for name, _ in self.items()]


def create_builtin_registry() -> BackendRegistry:
    registry = BackendRegistry()
    registry.register(CoreMLBackend())
    registry.register(ExecuTorchBackend())
    registry.register(LiteRTBackend())
    registry.register(NcnnBackend())
    registry.register(OnnxBackend())
    registry.register(TensorRTBackend())
    registry.register(TorchExportBackend())
    registry.register(TorchScriptBackend())
    return registry



from __future__ import annotations

from dataclasses import dataclass
from pathlib import Path
from typing import Annotated

import timm
import torch
import typer
from timm.data import resolve_data_config
from timm.utils import reparameterize_model

from timmx.errors import ConfigurationError, ExportError
from timmx.export.types import Device

# ---------------------------------------------------------------------------
# Typer type aliases for --normalize / --softmax CLI flags
# ---------------------------------------------------------------------------

NormalizeOpt = Annotated[
    bool,
    typer.Option(
        "--normalize",
        help="Wrap model with input normalization (mean/std from timm config).",
    ),
]
SoftmaxOpt = Annotated[
    bool,
    typer.Option(
        "--softmax",
        help="Add softmax output layer.",
    ),
]
MeanOpt = Annotated[
    tuple[float, float, float] | None,
    typer.Option(help="Custom normalization mean (3 RGB values; averaged for 1-channel models)."),
]
StdOpt = Annotated[
    tuple[float, float, float] | None,
    typer.Option(help="Custom normalization std (3 RGB values; averaged for 1-channel models)."),
]

DEFAULT_INPUT_SIZE = (3, 224, 224)
DEFAULT_MEAN = (0.485, 0.456, 0.406)
DEFAULT_STD = (0.229, 0.224, 0.225)
SUPPORTED_INPUT_CHANNELS = frozenset({1, 3})

# ---------------------------------------------------------------------------
# Shared Typer type aliases for common CLI parameters
# ---------------------------------------------------------------------------

ModelNameArg = Annotated[str, typer.Argument(help="timm model name, e.g. resnet18")]
OutputOpt = Annotated[Path, typer.Option(help="Path to write the exported model.")]
CheckpointOpt = Annotated[Path | None, typer.Option(help="Path to a fine-tuned checkpoint.")]
PretrainedOpt = Annotated[bool, typer.Option("--pretrained", help="Load timm pretrained weights.")]
NumClassesOpt = Annotated[
    int | None, typer.Option(help="Override the model classifier output classes.")
]
InChansOpt = Annotated[
    int | None,
    typer.Option(help="Override model input channels (currently 1 or 3 only)."),
]
BatchSizeOpt = Annotated[int, typer.Option(help="Example input batch size for export.")]
InputSizeOpt = Annotated[
    tuple[int, int, int] | None, typer.Option(help="Explicit input shape as C H W.")
]
DeviceOpt = Annotated[Device, typer.Option(help="Device used for model instantiation and export.")]


# ---------------------------------------------------------------------------
# PreparedExport: result of the common model-creation / input-preparation step
# ---------------------------------------------------------------------------


@dataclass
class PreparedExport:
    """Bundle returned by :func:`prepare_export` with everything backends need."""

    model: torch.nn.Module
    example_input: torch.Tensor
    resolved_input_size: tuple[int, int, int]
    output_path: Path
    torch_device: torch.device


def prepare_export(
    *,
    model_name: str,
    output: Path,
    checkpoint: Path | None,
    pretrained: bool,
    num_classes: int | None,
    in_chans: int | None,
    batch_size: int,
    input_size: tuple[int, int, int] | None,
    device: Device | str,
    output_is_dir: bool = False,
    normalize: bool = False,
    softmax: bool = False,
    mean: tuple[float, ...] | None = None,
    std: tuple[float, ...] | None = None,
) -> PreparedExport:
    """Validate common args, create the timm model, and build an example input.

    Set *output_is_dir=True* when the backend writes to a directory rather than a
    single file (e.g. ncnn).  The resolved path is then created as a directory;
    otherwise its parent directory is created.
    """
    validate_common_args(
        batch_size=batch_size,
        device=device,
        in_chans=in_chans,
        input_size=input_size,
    )

    if (mean is not None or std is not None) and not normalize:
        raise ConfigurationError("--mean/--std require --normalize.")

    output_path = Path(output).expanduser().resolve()
    if output_is_dir:
        output_path.mkdir(parents=True, exist_ok=True)
    else:
        output_path.parent.mkdir(parents=True, exist_ok=True)

    model = create_timm_model(
        model_name,
        pretrained=pretrained,
        checkpoint=checkpoint,
        num_classes=num_classes,
        in_chans=in_chans,
    )
    resolved_input_size = resolve_input_size(model, input_size)
    model_input_channels = resolve_model_input_channels(model)
    validate_supported_input_channels(model_input_channels, source="model")
    if resolved_input_size[0] != model_input_channels:
        raise ConfigurationError(
            "--input-size channel count must match the model input channels: "
            f"expected {model_input_channels}, got {resolved_input_size[0]}."
        )

    torch_device = torch.device(device)
    model = model.to(torch_device)
    model.eval()

    if softmax or normalize:
        model = wrap_with_preprocessing(
            model,
            normalize=normalize,
            softmax=softmax,
            mean=mean,
            std=std,
        )
        model = model.to(torch_device)

    example_input = torch.randn(batch_size, *resolved_input_size, device=torch_device)

    return PreparedExport(
        model=model,
        example_input=example_input,
        resolved_input_size=resolved_input_size,
        output_path=output_path,
        torch_device=torch_device,
    )


# ---------------------------------------------------------------------------
# Preprocessing / postprocessing wrapper
# ---------------------------------------------------------------------------


class PrePostWrapper(torch.nn.Module):
    """Wraps a model with optional input normalization and optional softmax output."""

    def __init__(
        self,
        model: torch.nn.Module,
        mean: tuple[float, ...] | None = None,
        std: tuple[float, ...] | None = None,
        normalize: bool = True,
        softmax: bool = False,
    ):
        super().__init__()
        self.model = model
        self.normalize = normalize
        if normalize:
            if mean is None or std is None:
                raise ConfigurationError("PrePostWrapper normalization requires mean/std values.")
            self.register_buffer("mean", torch.tensor(mean).reshape(1, -1, 1, 1))
            self.register_buffer("std", torch.tensor(std).reshape(1, -1, 1, 1))
        else:
            self.register_buffer("mean", None)
            self.register_buffer("std", None)
        self.softmax = softmax
        self.train(model.training)

    def forward(self, x: torch.Tensor) -> torch.Tensor:
        if self.normalize:
            x = (x - self.mean) / self.std
        x = self.model(x)
        if self.softmax:
            x = torch.nn.functional.softmax(x, dim=-1)
        return x


def wrap_with_preprocessing(
    model: torch.nn.Module,
    normalize: bool = True,
    softmax: bool = False,
    mean: tuple[float, ...] | None = None,
    std: tuple[float, ...] | None = None,
) -> PrePostWrapper:
    """Wrap model with optional normalization and optional softmax.

    Uses timm's data config for normalization by default; pass *mean*/*std* to override.
    """
    effective_mean: tuple[float, ...] | None
    effective_std: tuple[float, ...] | None
    if normalize:
        effective_mean, effective_std = resolve_normalization_stats(model, mean=mean, std=std)
    else:
        effective_mean, effective_std = None, None
    return PrePostWrapper(
        model,
        mean=effective_mean,
        std=effective_std,
        normalize=normalize,
        softmax=softmax,
    )


def resolve_model_input_channels(model: torch.nn.Module) -> int:
    raw_channels = getattr(model, "in_chans", None)
    if isinstance(raw_channels, int):
        return raw_channels

    config = resolve_data_config(model=model)
    raw_input_size = config.get("input_size")
    if isinstance(raw_input_size, tuple) and len(raw_input_size) == 3:
        return int(raw_input_size[0])
    if isinstance(raw_input_size, list) and len(raw_input_size) == 3:
        return int(raw_input_size[0])

    return DEFAULT_INPUT_SIZE[0]


def resolve_normalization_stats(
    model: torch.nn.Module,
    *,
    mean: tuple[float, ...] | None = None,
    std: tuple[float, ...] | None = None,
) -> tuple[tuple[float, ...], tuple[float, ...]]:
    config = resolve_data_config(model=model)
    input_channels = resolve_model_input_channels(model)
    return resolve_normalization_stats_for_channels(
        input_channels=input_channels,
        mean=mean,
        std=std,
        default_mean=config.get("mean", DEFAULT_MEAN),
        default_std=config.get("std", DEFAULT_STD),
    )


def resolve_normalization_stats_for_channels(
    *,
    input_channels: int,
    mean: tuple[float, ...] | None = None,
    std: tuple[float, ...] | None = None,
    default_mean: tuple[float, ...] | list[float] | None = DEFAULT_MEAN,
    default_std: tuple[float, ...] | list[float] | None = DEFAULT_STD,
) -> tuple[tuple[float, ...], tuple[float, ...]]:
    validate_supported_input_channels(input_channels, source="input")

    raw_mean = mean if mean is not None else (default_mean or DEFAULT_MEAN)
    raw_std = std if std is not None else (default_std or DEFAULT_STD)
    return (
        _coerce_stats_for_channels(raw_mean, input_channels, stat_name="mean"),
        _coerce_stats_for_channels(raw_std, input_channels, stat_name="std"),
    )


def validate_supported_input_channels(channels: int, *, source: str) -> None:
    if channels not in SUPPORTED_INPUT_CHANNELS:
        raise ConfigurationError(f"{source} input channels must be 1 or 3 for now, got {channels}.")


def _coerce_stats_for_channels(
    stats: tuple[float, ...] | list[float],
    input_channels: int,
    *,
    stat_name: str,
) -> tuple[float, ...]:
    values = tuple(float(v) for v in stats)
    if len(values) == input_channels:
        return values
    if len(values) == 3 and input_channels == 1:
        return (sum(values) / len(values),)
    if len(values) == 1 and input_channels == 3:
        return values * 3
    raise ConfigurationError(
        f"Normalization {stat_name} has {len(values)} values but only "
        "1-value and 3-value stats are supported."
    )


def create_timm_model(
    model_name: str,
    *,
    pretrained: bool,
    checkpoint: Path | None,
    num_classes: int | None,
    in_chans: int | None,
) -> torch.nn.Module:
    if in_chans is not None:
        validate_supported_input_channels(in_chans, source="--in-chans")

    create_kwargs: dict[str, object] = {
        "pretrained": pretrained,
        "exportable": True,
    }

    if checkpoint is not None:
        resolved_checkpoint = Path(checkpoint).expanduser()
        if not resolved_checkpoint.is_file():
            raise ConfigurationError(f"Checkpoint file does not exist: {resolved_checkpoint}")
        create_kwargs["checkpoint_path"] = str(resolved_checkpoint)

    if num_classes is not None:
        create_kwargs["num_classes"] = num_classes

    if in_chans is not None:
        create_kwargs["in_chans"] = in_chans

    try:
        model = timm.create_model(model_name, **create_kwargs)
    except Exception as exc:
        raise ExportError(f"Failed to create timm model {model_name!r}: {exc}") from exc

    reparameterize_model(model, inplace=True)
    return model


def resolve_input_size(
    model: torch.nn.Module,
    requested: tuple[int, int, int] | None,
) -> tuple[int, int, int]:
    if requested is not None:
        validate_supported_input_channels(requested[0], source="--input-size")
        return requested

    input_channels = resolve_model_input_channels(model)
    validate_supported_input_channels(input_channels, source="model")

    config = resolve_data_config(model=model)
    raw_input_size = config.get("input_size")
    if isinstance(raw_input_size, tuple) and len(raw_input_size) == 3:
        return input_channels, int(raw_input_size[1]), int(raw_input_size[2])
    if isinstance(raw_input_size, list) and len(raw_input_size) == 3:
        return input_channels, int(raw_input_size[1]), int(raw_input_size[2])

    return input_channels, DEFAULT_INPUT_SIZE[1], DEFAULT_INPUT_SIZE[2]


def validate_common_args(
    *,
    batch_size: int,
    device: str,
    in_chans: int | None = None,
    input_size: tuple[int, int, int] | None = None,
) -> None:
    if batch_size < 1:
        raise ConfigurationError("--batch-size must be >= 1.")
    if device == "cuda" and not torch.cuda.is_available():
        raise ConfigurationError("--device cuda was requested but CUDA is unavailable.")
    if in_chans is not None:
        validate_supported_input_channels(in_chans, source="--in-chans")
    if input_size is not None:
        validate_supported_input_channels(input_size[0], source="--input-size")



from __future__ import annotations

from collections.abc import Callable
from enum import StrEnum
from pathlib import Path
from typing import Annotated

import torch
import typer

from timmx.errors import ConfigurationError, ExportError
from timmx.export.base import DependencyStatus, ExportBackend
from timmx.export.calibration import resolve_calibration_batches
from timmx.export.common import (
    BatchSizeOpt,
    CheckpointOpt,
    DeviceOpt,
    InChansOpt,
    InputSizeOpt,
    MeanOpt,
    ModelNameArg,
    NormalizeOpt,
    NumClassesOpt,
    OutputOpt,
    PretrainedOpt,
    SoftmaxOpt,
    StdOpt,
    prepare_export,
)
from timmx.export.types import Device


class ExecuTorchDelegate(StrEnum):
    xnnpack = "xnnpack"
    coreml = "coreml"


class ExecuTorchMode(StrEnum):
    fp32 = "fp32"
    int8 = "int8"


class ComputePrecision(StrEnum):
    float16 = "float16"
    float32 = "float32"


class ExecuTorchBackend(ExportBackend):
    name = "executorch"
    help = "Export a timm model to ExecuTorch (.pte) with delegate acceleration."

    def check_dependencies(self) -> DependencyStatus:
        missing = []
        try:
            import executorch  # noqa: F401
        except ImportError:
            missing.append("executorch")
        return DependencyStatus(
            available=not missing,
            missing_packages=missing,
            install_hint="pip install 'timmx[executorch]'",
        )

    def create_command(self) -> Callable[..., None]:
        def command(
            model_name: ModelNameArg,
            output: OutputOpt,
            checkpoint: CheckpointOpt = None,
            pretrained: PretrainedOpt = False,
            num_classes: NumClassesOpt = None,
            in_chans: InChansOpt = None,
            batch_size: BatchSizeOpt = 1,
            input_size: InputSizeOpt = None,
            device: DeviceOpt = Device.cpu,
            delegate: Annotated[
                ExecuTorchDelegate,
                typer.Option(help="ExecuTorch delegate backend for hardware acceleration."),
            ] = ExecuTorchDelegate.xnnpack,
            mode: Annotated[
                ExecuTorchMode,
                typer.Option(help="Export precision mode."),
            ] = ExecuTorchMode.fp32,
            compute_precision: Annotated[
                ComputePrecision | None,
                typer.Option(help="CoreML compute precision. Only valid with --delegate coreml."),
            ] = None,
            dynamic_batch: Annotated[
                bool,
                typer.Option("--dynamic-batch", help="Mark batch axis as dynamic."),
            ] = False,
            calibration_data: Annotated[
                Path | None,
                typer.Option(
                    help=(
                        "Path to calibration data: an image directory or a "
                        "torch-saved tensor (N, C, H, W). Required for --mode int8 "
                        "unless --random-calibration is set."
                    )
                ),
            ] = None,
            calibration_steps: Annotated[
                int | None,
                typer.Option(
                    help=(
                        "Number of calibration batches to consume. "
                        "Default is all full batches from --calibration-data."
                    )
                ),
            ] = None,
            calibration_samples: Annotated[
                int | None,
                typer.Option(
                    help=(
                        "Max number of images to load when --calibration-data is a "
                        "directory. Default is 128."
                    )
                ),
            ] = None,
            random_calibration: Annotated[
                bool,
                typer.Option(
                    "--random-calibration",
                    help=(
                        "Use random noise for calibration instead of real data. "
                        "Not recommended for production use."
                    ),
                ),
            ] = False,
            per_channel: Annotated[
                bool,
                typer.Option(
                    help="Use per-channel quantization for int8. "
                    "Disable with --no-per-channel for per-tensor."
                ),
            ] = True,
            normalize: NormalizeOpt = False,
            softmax: SoftmaxOpt = False,
            mean: MeanOpt = None,
            std: StdOpt = None,
        ) -> None:
            if compute_precision is not None and delegate != ExecuTorchDelegate.coreml:
                raise ConfigurationError(
                    "--compute-precision is only supported with --delegate coreml."
                )
            if mode != ExecuTorchMode.int8 and (
                calibration_data is not None
                or calibration_steps is not None
                or calibration_samples is not None
                or random_calibration
            ):
                raise ConfigurationError(
                    "--calibration-data, --calibration-steps, --calibration-samples, "
                    "and --random-calibration are only valid with --mode int8."
                )
            if dynamic_batch and batch_size < 2:
                raise ConfigurationError(
                    "--dynamic-batch requires --batch-size >= 2 for stable symbolic shape capture."
                )
            if (mean is not None or std is not None) and not normalize:
                if mode != ExecuTorchMode.int8:
                    raise ConfigurationError(
                        "--mean/--std require --normalize unless used for --mode int8 calibration."
                    )

            _import_executorch()

            prep = prepare_export(
                model_name=model_name,
                output=output,
                checkpoint=checkpoint,
                pretrained=pretrained,
                num_classes=num_classes,
                in_chans=in_chans,
                batch_size=batch_size,
                input_size=input_size,
                device=device,
                normalize=normalize,
                softmax=softmax,
                mean=mean if normalize else None,
                std=std if normalize else None,
            )

            partitioner = _build_partitioner(
                delegate=delegate,
                compute_precision=compute_precision,
                quantized=mode == ExecuTorchMode.int8,
            )

            if mode == ExecuTorchMode.int8:
                et_program = _export_quantized(
                    model=prep.model,
                    example_input=prep.example_input,
                    resolved_input_size=prep.resolved_input_size,
                    torch_device=prep.torch_device,
                    batch_size=batch_size,
                    calibration_data=calibration_data,
                    calibration_steps=calibration_steps,
                    calibration_samples=calibration_samples,
                    random_calibration=random_calibration,
                    per_channel=per_channel,
                    dynamic_batch=dynamic_batch,
                    delegate=delegate,
                    partitioner=partitioner,
                    mean=mean,
                    std=std,
                    normalize_calibration_images=not normalize,
                )
            else:
                et_program = _export_standard(
                    model=prep.model,
                    example_input=prep.example_input,
                    dynamic_batch=dynamic_batch,
                    partitioner=partitioner,
                )

            try:
                with open(prep.output_path, "wb") as f:
                    et_program.write_to_file(f)
            except Exception as exc:
                raise ExportError(
                    f"Failed to write ExecuTorch model to {prep.output_path}: {exc}"
                ) from exc

        return command


def _export_standard(
    *,
    model: torch.nn.Module,
    example_input: torch.Tensor,
    dynamic_batch: bool,
    partitioner: list[object],
) -> object:
    from executorch.exir import to_edge_transform_and_lower

    dynamic_shapes: tuple[dict[int, torch.export.Dim], ...] | None = None
    if dynamic_batch:
        dynamic_shapes = ({0: torch.export.Dim("batch")},)

    try:
        exported_program = torch.export.export(
            model,
            (example_input,),
            dynamic_shapes=dynamic_shapes,
        )
    except Exception as exc:
        raise ExportError(f"torch.export capture failed: {exc}") from exc

    try:
        et_program = to_edge_transform_and_lower(
            exported_program,
            partitioner=partitioner,
        ).to_executorch()
    except Exception as exc:
        raise ExportError(f"ExecuTorch edge lowering failed: {exc}") from exc

    return et_program


def _export_quantized(
    *,
    model: torch.nn.Module,
    example_input: torch.Tensor,
    resolved_input_size: tuple[int, int, int],
    torch_device: torch.device,
    batch_size: int,
    calibration_data: Path | None,
    calibration_steps: int | None,
    calibration_samples: int | None,
    random_calibration: bool,
    per_channel: bool,
    dynamic_batch: bool,
    delegate: ExecuTorchDelegate,
    partitioner: list[object],
    mean: tuple[float, ...] | None = None,
    std: tuple[float, ...] | None = None,
    normalize_calibration_images: bool = True,
) -> object:
    from executorch.exir import EdgeCompileConfig, to_edge_transform_and_lower
    from torchao.quantization.pt2e import quantize_pt2e

    calibration_batches = resolve_calibration_batches(
        calibration_data=calibration_data,
        calibration_steps=calibration_steps,
        batch_size=batch_size,
        input_size=resolved_input_size,
        device=torch_device,
        model=model,
        calibration_samples=calibration_samples,
        random_calibration=random_calibration,
        mean=mean,
        std=std,
        normalize_images=normalize_calibration_images,
    )

    quantizer = _build_quantizer(delegate=delegate, per_channel=per_channel)

    try:
        exported_module = torch.export.export(model, (example_input,)).module()

        prepared = quantize_pt2e.prepare_pt2e(exported_module, quantizer)

        with torch.no_grad():
            for batch in calibration_batches:
                prepared(batch)

        quantized = quantize_pt2e.convert_pt2e(prepared)
    except Exception as exc:
        raise ExportError(f"PT2E quantization failed: {exc}") from exc

    dynamic_shapes: tuple[dict[int, torch.export.Dim], ...] | None = None
    if dynamic_batch:
        dynamic_shapes = ({0: torch.export.Dim("batch")},)

    try:
        exported_program = torch.export.export(
            quantized,
            (example_input,),
            dynamic_shapes=dynamic_shapes,
        )
    except Exception as exc:
        raise ExportError(f"torch.export capture of quantized model failed: {exc}") from exc

    try:
        et_program = to_edge_transform_and_lower(
            exported_program,
            compile_config=EdgeCompileConfig(_check_ir_validity=False),
            partitioner=partitioner,
        ).to_executorch()
    except Exception as exc:
        raise ExportError(f"ExecuTorch edge lowering of quantized model failed: {exc}") from exc

    return et_program


def _build_quantizer(*, delegate: ExecuTorchDelegate, per_channel: bool) -> object:
    if delegate == ExecuTorchDelegate.xnnpack:
        try:
            from executorch.backends.xnnpack.quantizer.xnnpack_quantizer import (
                XNNPACKQuantizer,
                get_symmetric_quantization_config,
            )
        except ImportError as exc:
            raise ExportError(
                "XNNPACKQuantizer is required for --mode int8 --delegate xnnpack. "
                "Install with: pip install 'timmx[executorch]'"
            ) from exc

        quantizer = XNNPACKQuantizer()
        quantizer.set_global(
            get_symmetric_quantization_config(
                is_per_channel=per_channel,
                is_dynamic=False,
            )
        )
        return quantizer

    if delegate == ExecuTorchDelegate.coreml:
        try:
            from coremltools.optimize.torch.quantization import LinearQuantizerConfig
            from executorch.backends.apple.coreml.quantizer import CoreMLQuantizer
        except ImportError as exc:
            raise ExportError(
                "CoreMLQuantizer and coremltools are required for --mode int8 --delegate coreml. "
                "Install with: pip install 'timmx[executorch]' coremltools"
            ) from exc

        config = LinearQuantizerConfig.from_dict(
            {
                "global_config": {
                    "weight_dtype": "qint8",
                    "activation_dtype": "quint8",
                    "weight_per_channel": per_channel,
                }
            }
        )
        return CoreMLQuantizer(config)

    raise ConfigurationError(f"Unknown delegate: {delegate}")


def _build_partitioner(
    *,
    delegate: ExecuTorchDelegate,
    compute_precision: ComputePrecision | None,
    quantized: bool = False,
) -> list[object]:
    if delegate == ExecuTorchDelegate.xnnpack:
        try:
            from executorch.backends.xnnpack.partition.xnnpack_partitioner import (
                XnnpackPartitioner,
            )
        except ImportError as exc:
            raise ExportError(
                "XnnpackPartitioner is required for --delegate xnnpack. "
                "Install with: pip install 'timmx[executorch]'"
            ) from exc
        return [XnnpackPartitioner()]

    if delegate == ExecuTorchDelegate.coreml:
        try:
            from executorch.backends.apple.coreml.partition import CoreMLPartitioner
        except ImportError as exc:
            raise ExportError(
                "CoreMLPartitioner is required for --delegate coreml. "
                "Ensure executorch is installed with CoreML support (macOS only)."
            ) from exc

        compile_specs = _build_coreml_compile_specs(compute_precision, quantized=quantized)
        if compile_specs is not None:
            return [CoreMLPartitioner(compile_specs=compile_specs)]
        return [CoreMLPartitioner()]

    raise ConfigurationError(f"Unknown delegate: {delegate}")


def _build_coreml_compile_specs(
    compute_precision: ComputePrecision | None,
    *,
    quantized: bool = False,
) -> list[object] | None:
    if compute_precision is None and not quantized:
        return None

    import coremltools as ct
    from executorch.backends.apple.coreml.compiler import CoreMLBackend

    kwargs: dict[str, object] = {}
    if compute_precision is not None:
        kwargs["compute_precision"] = (
            ct.precision.FLOAT32
            if compute_precision == ComputePrecision.float32
            else ct.precision.FLOAT16
        )
    if quantized:
        kwargs["minimum_deployment_target"] = ct.target.iOS17
    return CoreMLBackend.generate_compile_specs(**kwargs)


def _import_executorch() -> None:
    try:
        import executorch  # noqa: F401
    except ImportError as exc:
        raise ExportError(
            "executorch is required for ExecuTorch export. "
            "Install with: pip install 'timmx[executorch]'"
        ) from exc



from __future__ import annotations

from collections.abc import Callable
from typing import Annotated

import torch
import typer

from timmx.errors import ConfigurationError, ExportError
from timmx.export.base import DependencyStatus, ExportBackend
from timmx.export.common import (
    BatchSizeOpt,
    CheckpointOpt,
    DeviceOpt,
    InChansOpt,
    InputSizeOpt,
    MeanOpt,
    ModelNameArg,
    NormalizeOpt,
    NumClassesOpt,
    OutputOpt,
    PretrainedOpt,
    SoftmaxOpt,
    StdOpt,
    prepare_export,
)
from timmx.export.types import Device

DEFAULT_OPSET = 18


class OnnxBackend(ExportBackend):
    name = "onnx"
    help = "Export a timm model to ONNX."

    def check_dependencies(self) -> DependencyStatus:
        missing = []
        for mod in ("onnx", "onnxscript", "onnxslim"):
            try:
                __import__(mod)
            except ImportError:
                missing.append(mod)
        return DependencyStatus(
            available=not missing,
            missing_packages=missing,
            install_hint="pip install 'timmx[onnx]'",
        )

    def create_command(self) -> Callable[..., None]:
        def command(
            model_name: ModelNameArg,
            output: OutputOpt,
            checkpoint: CheckpointOpt = None,
            pretrained: PretrainedOpt = False,
            num_classes: NumClassesOpt = None,
            in_chans: InChansOpt = None,
            batch_size: BatchSizeOpt = 1,
            input_size: InputSizeOpt = None,
            opset: Annotated[
                int, typer.Option(help="ONNX opset version to target.")
            ] = DEFAULT_OPSET,
            dynamic_batch: Annotated[
                bool, typer.Option("--dynamic-batch", help="Mark batch axis as dynamic.")
            ] = False,
            device: DeviceOpt = Device.cpu,
            external_data: Annotated[
                bool, typer.Option(help="Save large model weights in external data files.")
            ] = False,
            check: Annotated[bool, typer.Option(help="Run ONNX checker after export.")] = True,
            slim: Annotated[
                bool, typer.Option(help="Optimize the exported model with onnxslim.")
            ] = True,
            normalize: NormalizeOpt = False,
            softmax: SoftmaxOpt = False,
            mean: MeanOpt = None,
            std: StdOpt = None,
        ) -> None:
            if opset < 7:
                raise ConfigurationError("--opset must be >= 7.")

            prep = prepare_export(
                model_name=model_name,
                output=output,
                checkpoint=checkpoint,
                pretrained=pretrained,
                num_classes=num_classes,
                in_chans=in_chans,
                batch_size=batch_size,
                input_size=input_size,
                device=device,
                normalize=normalize,
                softmax=softmax,
                mean=mean,
                std=std,
            )

            export_kwargs: dict[str, object] = {
                "f": str(prep.output_path),
                "opset_version": opset,
                "input_names": ["input"],
                "output_names": ["output"],
                "dynamo": True,
                "fallback": True,
                "external_data": external_data,
            }
            if dynamic_batch:
                export_kwargs["dynamic_shapes"] = ({0: torch.export.Dim("batch")},)
                export_kwargs["dynamic_axes"] = {"input": {0: "batch"}, "output": {0: "batch"}}

            try:
                torch.onnx.export(prep.model, (prep.example_input,), **export_kwargs)
            except Exception as exc:
                raise ExportError(f"ONNX export failed: {exc}") from exc

            if slim:
                try:
                    import onnxslim
                except ImportError as exc:
                    raise ExportError(
                        "onnxslim is required for ONNX model optimization. "
                        "Install with: pip install 'timmx[onnx]'"
                    ) from exc

                try:
                    onnxslim.slim(str(prep.output_path), str(prep.output_path))
                except Exception as exc:
                    raise ExportError(f"onnxslim optimization failed: {exc}") from exc

            if check:
                try:
                    import onnx
                except ImportError as exc:
                    raise ExportError(
                        "onnx is required for ONNX model checking. "
                        "Install with: pip install 'timmx[onnx]'"
                    ) from exc

                try:
                    onnx.checker.check_model(str(prep.output_path))
                except Exception as exc:
                    raise ExportError(f"Exported model failed ONNX check: {exc}") from exc

        return command



from timmx.export.base import DependencyStatus
from timmx.export.registry import BackendRegistry, create_builtin_registry

__all__ = ["BackendRegistry", "DependencyStatus", "create_builtin_registry"]



from __future__ import annotations

import random
from pathlib import Path

import torch

from timmx.console import console
from timmx.errors import ConfigurationError, ExportError
from timmx.export.common import (
    resolve_normalization_stats_for_channels,
    validate_supported_input_channels,
)

IMAGE_EXTENSIONS = {".jpg", ".jpeg", ".png", ".bmp", ".webp", ".tiff", ".tif"}
DEFAULT_CALIBRATION_SAMPLES = 128


def resolve_calibration_batches(
    *,
    calibration_data: Path | None,
    calibration_steps: int | None,
    batch_size: int,
    input_size: tuple[int, int, int],
    device: torch.device,
    model: torch.nn.Module | None = None,
    calibration_samples: int | None = None,
    random_calibration: bool = False,
    mean: tuple[float, ...] | None = None,
    std: tuple[float, ...] | None = None,
    normalize_images: bool = True,
) -> list[torch.Tensor]:
    validate_supported_input_channels(input_size[0], source="calibration input")

    if calibration_steps is not None and calibration_steps < 1:
        raise ConfigurationError("--calibration-steps must be >= 1.")

    if calibration_samples is not None and calibration_samples < 1:
        raise ConfigurationError("--calibration-samples must be >= 1.")

    if calibration_data is None:
        if not random_calibration:
            raise ConfigurationError(
                "int8 quantization requires calibration data for accurate results.\n\n"
                "  Provide real data:      --calibration-data <image-dir-or-tensor.pt>\n"
                "  Or use random noise:    --random-calibration\n"
                "  Or choose a different quantization mode (e.g. fp16, dynamic-int8).\n\n"
                "Random noise calibration produces poor quantization accuracy and "
                "is not recommended for production use."
            )
        console.print(
            "[bold yellow]Warning:[/bold yellow] Using random noise for calibration. "
            "For better quantization accuracy, provide real data via "
            "[bold]--calibration-data <image-dir-or-tensor.pt>[/bold]."
        )
        batch_count = calibration_steps or 1
        return [torch.randn(batch_size, *input_size, device=device) for _ in range(batch_count)]

    resolved_path = calibration_data.expanduser().resolve()

    if resolved_path.is_dir():
        if model is None:
            raise ConfigurationError(
                "model is required when loading calibration images from a directory."
            )
        data_tensor = _load_calibration_images(
            image_dir=resolved_path,
            model=model,
            input_size=input_size,
            max_samples=calibration_samples or DEFAULT_CALIBRATION_SAMPLES,
            mean=mean,
            std=std,
            normalize_images=normalize_images,
        )
    elif resolved_path.is_file():
        data_tensor = _load_calibration_tensor(resolved_path)
        data_tensor = _normalize_calibration_tensor(
            data_tensor=data_tensor,
            input_size=input_size,
            source_path=resolved_path,
        )
    else:
        raise ConfigurationError(f"Calibration data path does not exist: {resolved_path}")

    return _slice_into_batches(
        data_tensor=data_tensor,
        batch_size=batch_size,
        calibration_steps=calibration_steps,
        device=device,
        source_path=resolved_path,
    )


def _slice_into_batches(
    *,
    data_tensor: torch.Tensor,
    batch_size: int,
    calibration_steps: int | None,
    device: torch.device,
    source_path: Path,
) -> list[torch.Tensor]:
    full_batches = data_tensor.shape[0] // batch_size
    if full_batches < 1:
        raise ConfigurationError(
            f"Calibration data {source_path} has {data_tensor.shape[0]} samples but "
            f"needs at least --batch-size {batch_size}."
        )

    requested_batches = full_batches if calibration_steps is None else calibration_steps
    if requested_batches > full_batches:
        raise ConfigurationError(
            f"Calibration data {source_path} has only {full_batches} full batches for "
            f"--batch-size {batch_size}, but --calibration-steps is {requested_batches}."
        )

    batches: list[torch.Tensor] = []
    for batch_index in range(requested_batches):
        start = batch_index * batch_size
        stop = start + batch_size
        batch = data_tensor[start:stop].to(device=device, dtype=torch.float32)
        batches.append(batch)

    return batches


# ---------------------------------------------------------------------------
# Image directory loading
# ---------------------------------------------------------------------------


def _collect_image_paths(image_dir: Path) -> list[Path]:
    """Recursively collect image file paths from a directory."""
    paths = [
        p
        for p in sorted(image_dir.rglob("*"))
        if p.is_file() and p.suffix.lower() in IMAGE_EXTENSIONS
    ]
    if not paths:
        raise ConfigurationError(
            f"No image files found in {image_dir}. "
            f"Supported formats: {', '.join(sorted(IMAGE_EXTENSIONS))}"
        )
    return paths


def _load_calibration_images(
    *,
    image_dir: Path,
    model: torch.nn.Module,
    input_size: tuple[int, int, int],
    max_samples: int,
    mean: tuple[float, ...] | None = None,
    std: tuple[float, ...] | None = None,
    normalize_images: bool = True,
) -> torch.Tensor:
    from PIL import Image
    from timm.data import create_transform, resolve_data_config

    image_paths = _collect_image_paths(image_dir)

    if len(image_paths) > max_samples:
        rng = random.Random(42)
        image_paths = rng.sample(image_paths, max_samples)

    data_config = resolve_data_config(model=model)
    data_config["input_size"] = input_size
    if normalize_images:
        data_config["mean"], data_config["std"] = resolve_normalization_stats_for_channels(
            input_channels=input_size[0],
            mean=mean,
            std=std,
            default_mean=data_config.get("mean"),
            default_std=data_config.get("std"),
        )
    else:
        # When the exported model already embeds normalization, calibration images should
        # stay in the raw [0, 1] input space.
        data_config["mean"], data_config["std"] = resolve_normalization_stats_for_channels(
            input_channels=input_size[0],
            mean=(0.0, 0.0, 0.0),
            std=(1.0, 1.0, 1.0),
            default_mean=(0.0, 0.0, 0.0),
            default_std=(1.0, 1.0, 1.0),
        )
    transform = create_transform(**data_config, is_training=False)
    image_mode = "L" if input_size[0] == 1 else "RGB"

    tensors: list[torch.Tensor] = []
    skipped = 0
    for path in image_paths:
        try:
            img = Image.open(path).convert(image_mode)
            tensor = transform(img)
            tensors.append(tensor)
        except Exception as exc:
            skipped += 1
            console.print(f"[dim]  Skipped {path.name}: {exc}[/dim]")
            continue

    if not tensors:
        raise ConfigurationError(
            f"Failed to load any images from {image_dir}. "
            "Check that the directory contains valid image files."
        )

    if skipped > 0:
        console.print(f"[yellow]Skipped {skipped} unreadable image(s) from {image_dir}.[/yellow]")

    console.print(
        f"Loaded [bold]{len(tensors)}[/bold] calibration image(s) from [bold]{image_dir}[/bold]."
    )

    return torch.stack(tensors)


# ---------------------------------------------------------------------------
# Torch tensor file loading (existing path)
# ---------------------------------------------------------------------------


def _load_calibration_tensor(resolved_path: Path) -> torch.Tensor:
    try:
        payload = torch.load(resolved_path, map_location="cpu", weights_only=False)
    except Exception as exc:
        raise ExportError(f"Failed to load calibration data from {resolved_path}: {exc}") from exc

    if not torch.is_tensor(payload):
        raise ConfigurationError(
            f"Calibration data at {resolved_path} must be a torch.Tensor with shape (N, C, H, W)."
        )

    return payload


def _normalize_calibration_tensor(
    *,
    data_tensor: torch.Tensor,
    input_size: tuple[int, int, int],
    source_path: Path,
) -> torch.Tensor:
    if data_tensor.ndim == 3:
        data_tensor = data_tensor.unsqueeze(0)

    if data_tensor.ndim != 4:
        raise ConfigurationError(
            f"Calibration data {source_path} must have rank 4 (N, C, H, W), got rank "
            f"{data_tensor.ndim}."
        )

    channels, height, width = (
        int(data_tensor.shape[1]),
        int(data_tensor.shape[2]),
        int(data_tensor.shape[3]),
    )

    if (channels, height, width) != input_size:
        raise ConfigurationError(
            f"Calibration data {source_path} has per-sample shape "
            f"({channels}, {height}, {width}) but expected {input_size}."
        )

    if not data_tensor.is_floating_point():
        data_tensor = data_tensor.to(dtype=torch.float32)
    else:
        data_tensor = data_tensor.detach().to(dtype=torch.float32)

    return data_tensor.contiguous()



from __future__ import annotations

import functools
import logging
import os
import warnings

import typer

from timmx import __version__
from timmx.console import console
from timmx.errors import TimmxError
from timmx.export import BackendRegistry, DependencyStatus, create_builtin_registry
from timmx.export.base import ExportBackend
from timmx.export.common import (
    CheckpointOpt,
    InChansOpt,
    ModelNameArg,
    NumClassesOpt,
    PretrainedOpt,
)

app = typer.Typer(
    name="timmx", help="Export timm models to deployment formats.", no_args_is_help=True
)
export_app = typer.Typer(help="Export a model to a deployment format.", no_args_is_help=True)
app.add_typer(export_app, name="export")


def _version_callback(value: bool) -> None:
    if value:
        console.print(f"timmx {__version__}")
        raise typer.Exit()


@app.callback()
def _root_callback(
    version: bool = typer.Option(
        False,
        "--version",
        help="Show version and exit.",
        callback=_version_callback,
        is_eager=True,
    ),
) -> None:
    pass


def build_export_app(registry: BackendRegistry | None = None) -> None:
    active_registry = registry or create_builtin_registry()

    for name, backend in active_registry.items():
        command_fn = backend.create_command()

        @functools.wraps(command_fn)
        def wrapped(*args: object, _fn: object = command_fn, **kwargs: object) -> None:
            try:
                _fn(*args, **kwargs)
            except TimmxError as exc:
                console.print(f"[bold red]error:[/bold red] {exc}", highlight=False)
                raise typer.Exit(code=2) from exc

        export_app.command(name=name, help=backend.help)(wrapped)


build_export_app()


def _quiet_check(backend: ExportBackend) -> DependencyStatus:
    """Run check_dependencies with noisy third-party warnings suppressed."""
    prev_tf = os.environ.get("TF_CPP_MIN_LOG_LEVEL")
    os.environ["TF_CPP_MIN_LOG_LEVEL"] = "3"
    prev_level = logging.root.level
    logging.root.setLevel(logging.ERROR)
    try:
        with warnings.catch_warnings():
            warnings.simplefilter("ignore")
            return backend.check_dependencies()
    finally:
        logging.root.setLevel(prev_level)
        if prev_tf is None:
            os.environ.pop("TF_CPP_MIN_LOG_LEVEL", None)
        else:
            os.environ["TF_CPP_MIN_LOG_LEVEL"] = prev_tf


@app.command()
def doctor() -> None:
    """Check timmx installation and backend availability."""
    import platform

    import torch
    from rich.table import Table

    console.print(f"[bold]timmx[/bold] v{__version__}")
    console.print(f"Python {platform.python_version()} | torch {torch.__version__}")
    console.print()

    table = Table(title="Backends", show_lines=False)
    table.add_column("Backend", style="bold")
    table.add_column("Status")
    table.add_column("Install")

    registry = create_builtin_registry()
    all_available = True

    for name, backend in registry.items():
        status = _quiet_check(backend)
        if status.available:
            table.add_row(name, "[green]:white_check_mark: available[/green]", "")
        else:
            all_available = False
            missing_str = ", ".join(status.missing_packages)
            table.add_row(
                name,
                f"[red]:x: missing[/red] ({missing_str})",
                "[dim]{}[/dim]".format(status.install_hint.replace("[", r"\[")),
            )

    console.print(table)

    if not all_available:
        console.print()
        console.print(
            "[dim]Tip: pip install 'timmx\\[all]' installs all"
            " non-platform-specific backends.[/dim]"
        )


def _format_param_count(n: int) -> str:
    """Format a parameter count like '11.69M (11,689,512)'."""
    if n >= 1_000_000:
        return f"{n / 1_000_000:.2f}M ({n:,})"
    if n >= 1_000:
        return f"{n / 1_000:.2f}K ({n:,})"
    return f"{n:,}"


@app.command()
def info(
    model_name: ModelNameArg,
    pretrained: PretrainedOpt = False,
    checkpoint: CheckpointOpt = None,
    num_classes: NumClassesOpt = None,
    in_chans: InChansOpt = None,
) -> None:
    """Show model metadata without exporting."""
    from rich.table import Table

    from timmx.export.common import create_timm_model, resolve_input_size

    try:
        model = create_timm_model(
            model_name,
            pretrained=pretrained,
            checkpoint=checkpoint,
            num_classes=num_classes,
            in_chans=in_chans,
        )
    except TimmxError as exc:
        console.print(f"[bold red]error:[/bold red] {exc}", highlight=False)
        raise typer.Exit(code=2) from exc

    input_size = resolve_input_size(model, None)
    total_params = sum(p.numel() for p in model.parameters())
    trainable_params = sum(p.numel() for p in model.parameters() if p.requires_grad)

    # Determine weights status
    if checkpoint is not None:
        weights_info = f"checkpoint ({checkpoint.name})"
    elif pretrained:
        weights_info = "pretrained"
    else:
        weights_info = "none (random init)"

    table = Table(show_header=False, show_edge=False, pad_edge=False)
    table.add_column(style="bold")
    table.add_column()

    table.add_row("Model", model_name)
    table.add_row("Architecture", model.__class__.__name__)
    table.add_row("Parameters", _format_param_count(total_params))
    if trainable_params != total_params:
        table.add_row("Trainable", _format_param_count(trainable_params))
    table.add_row("Input size", f"{input_size[0]} x {input_size[1]} x {input_size[2]}")
    num_cls = getattr(model, "num_classes", None)
    if num_cls is not None:
        table.add_row("Classes", str(num_cls))
    num_feat = getattr(model, "num_features", None)
    if num_feat is not None:
        table.add_row("Feature dim", str(num_feat))
    table.add_row("Weights", weights_info)

    console.print(table)


@app.command(name="list")
def list_models(
    query: str = typer.Argument("", help="Filter pattern (substring or glob)."),
    pretrained_only: bool = typer.Option(
        False, "--pretrained-only", help="Only show models with pretrained weights."
    ),
) -> None:
    """List available timm models."""
    import timm

    pattern = query if any(c in query for c in "*?[") else f"*{query}*"
    models = timm.list_models(pattern, pretrained=pretrained_only)
    for name in models:
        console.print(name, highlight=False)
    console.print()
    label = "model" if len(models) == 1 else "models"
    console.print(f"Found {len(models)} {label}")


def main() -> None:
    app()



from __future__ import annotations

from collections.abc import Callable
from enum import StrEnum
from typing import Annotated

import torch
import typer

from timmx.errors import ExportError
from timmx.export.base import ExportBackend
from timmx.export.common import (
    BatchSizeOpt,
    CheckpointOpt,
    DeviceOpt,
    InChansOpt,
    InputSizeOpt,
    MeanOpt,
    ModelNameArg,
    NormalizeOpt,
    NumClassesOpt,
    OutputOpt,
    PretrainedOpt,
    SoftmaxOpt,
    StdOpt,
    prepare_export,
)
from timmx.export.types import Device


class ScriptMethod(StrEnum):
    trace = "trace"
    script = "script"


class TorchScriptBackend(ExportBackend):
    name = "torchscript"
    help = "Export a timm model to TorchScript (.pt)."

    def create_command(self) -> Callable[..., None]:
        def command(
            model_name: ModelNameArg,
            output: OutputOpt,
            checkpoint: CheckpointOpt = None,
            pretrained: PretrainedOpt = False,
            num_classes: NumClassesOpt = None,
            in_chans: InChansOpt = None,
            batch_size: BatchSizeOpt = 1,
            input_size: InputSizeOpt = None,
            method: Annotated[
                ScriptMethod,
                typer.Option(help="Scripting method: trace (recommended) or script."),
            ] = ScriptMethod.trace,
            device: DeviceOpt = Device.cpu,
            verify: Annotated[
                bool,
                typer.Option(help="Load the saved model after export and run a forward pass."),
            ] = True,
            normalize: NormalizeOpt = False,
            softmax: SoftmaxOpt = False,
            mean: MeanOpt = None,
            std: StdOpt = None,
        ) -> None:
            prep = prepare_export(
                model_name=model_name,
                output=output,
                checkpoint=checkpoint,
                pretrained=pretrained,
                num_classes=num_classes,
                in_chans=in_chans,
                batch_size=batch_size,
                input_size=input_size,
                device=device,
                normalize=normalize,
                softmax=softmax,
                mean=mean,
                std=std,
            )

            try:
                if method == ScriptMethod.trace:
                    scripted = torch.jit.trace(prep.model, prep.example_input)
                else:
                    scripted = torch.jit.script(prep.model)
            except Exception as exc:
                raise ExportError(f"TorchScript {method} failed: {exc}") from exc

            try:
                torch.jit.save(scripted, str(prep.output_path))
            except Exception as exc:
                raise ExportError(f"Failed to save TorchScript model: {exc}") from exc

            if verify:
                try:
                    loaded = torch.jit.load(str(prep.output_path), map_location=prep.torch_device)
                    loaded(prep.example_input)
                except Exception as exc:
                    raise ExportError(
                        f"Saved TorchScript model failed verification: {exc}"
                    ) from exc

        return command



class TimmxError(Exception):
    """Base exception for user-facing timmx errors."""


class ConfigurationError(TimmxError):
    """Raised when CLI arguments are invalid or incompatible."""


class ExportError(TimmxError):
    """Raised when an export operation fails."""



from __future__ import annotations

import shutil
from collections.abc import Callable
from pathlib import Path
from typing import Annotated

import typer

from timmx.errors import ExportError
from timmx.export.base import DependencyStatus, ExportBackend
from timmx.export.common import (
    BatchSizeOpt,
    CheckpointOpt,
    DeviceOpt,
    InChansOpt,
    InputSizeOpt,
    MeanOpt,
    ModelNameArg,
    NormalizeOpt,
    NumClassesOpt,
    PretrainedOpt,
    SoftmaxOpt,
    StdOpt,
    prepare_export,
)
from timmx.export.types import Device


class NcnnBackend(ExportBackend):
    name = "ncnn"
    help = "Export a timm model to ncnn (.param/.bin) via pnnx."

    def check_dependencies(self) -> DependencyStatus:
        try:
            __import__("pnnx")
            return DependencyStatus(available=True, missing_packages=[], install_hint="")
        except ImportError:
            return DependencyStatus(
                available=False,
                missing_packages=["pnnx"],
                install_hint="pip install 'timmx[ncnn]'",
            )

    def create_command(self) -> Callable[..., None]:
        def command(
            model_name: ModelNameArg,
            output: Annotated[
                Path, typer.Option(help="Output directory to write the exported ncnn model files.")
            ],
            checkpoint: CheckpointOpt = None,
            pretrained: PretrainedOpt = False,
            num_classes: NumClassesOpt = None,
            in_chans: InChansOpt = None,
            batch_size: BatchSizeOpt = 1,
            input_size: InputSizeOpt = None,
            device: DeviceOpt = Device.cpu,
            fp16: Annotated[
                bool,
                typer.Option(help="Export weights in fp16 precision."),
            ] = True,
            normalize: NormalizeOpt = False,
            softmax: SoftmaxOpt = False,
            mean: MeanOpt = None,
            std: StdOpt = None,
        ) -> None:
            prep = prepare_export(
                model_name=model_name,
                output=output,
                checkpoint=checkpoint,
                pretrained=pretrained,
                num_classes=num_classes,
                in_chans=in_chans,
                batch_size=batch_size,
                input_size=input_size,
                device=device,
                output_is_dir=True,
                normalize=normalize,
                softmax=softmax,
                mean=mean,
                std=std,
            )
            output_dir = prep.output_path

            try:
                import pnnx
            except ImportError as exc:
                raise ExportError(
                    "pnnx is required for ncnn export. Install with: pip install 'timmx[ncnn]'"
                ) from exc

            ptpath = output_dir / "model.pt"
            pnnxparam = output_dir / "model.pnnx.param"
            pnnxbin = output_dir / "model.pnnx.bin"
            pnnxpy = output_dir / "model_pnnx.py"
            pnnxonnx = output_dir / "model.pnnx.onnx"
            ncnnparam = output_dir / "model.ncnn.param"
            ncnnbin = output_dir / "model.ncnn.bin"
            ncnnpy = output_dir / "model_ncnn.py"

            try:
                pnnx.export(
                    prep.model,
                    inputs=prep.example_input,
                    ptpath=str(ptpath),
                    pnnxparam=str(pnnxparam),
                    pnnxbin=str(pnnxbin),
                    pnnxpy=str(pnnxpy),
                    pnnxonnx=str(pnnxonnx),
                    ncnnparam=str(ncnnparam),
                    ncnnbin=str(ncnnbin),
                    ncnnpy=str(ncnnpy),
                    fp16=fp16,
                    device=str(device),
                )
            except Exception as exc:
                raise ExportError(f"ncnn export via pnnx failed: {exc}") from exc

            # Remove pnnx intermediate files and the torchscript checkpoint
            for path in (ptpath, pnnxparam, pnnxbin, pnnxpy, pnnxonnx):
                path.unlink(missing_ok=True)

            # Remove __pycache__ that pnnx writes alongside the Python wrappers
            pycache = output_dir / "__pycache__"
            if pycache.exists():
                shutil.rmtree(pycache, ignore_errors=True)

        return command



from __future__ import annotations

from collections.abc import Callable
from typing import Annotated

import torch
import typer

from timmx.errors import ConfigurationError, ExportError
from timmx.export.base import ExportBackend
from timmx.export.common import (
    BatchSizeOpt,
    CheckpointOpt,
    DeviceOpt,
    InChansOpt,
    InputSizeOpt,
    MeanOpt,
    ModelNameArg,
    NormalizeOpt,
    NumClassesOpt,
    OutputOpt,
    PretrainedOpt,
    SoftmaxOpt,
    StdOpt,
    prepare_export,
)
from timmx.export.types import Device


class TorchExportBackend(ExportBackend):
    name = "torch-export"
    help = "Export a timm model with torch.export (.pt2)."

    def create_command(self) -> Callable[..., None]:
        def command(
            model_name: ModelNameArg,
            output: OutputOpt,
            checkpoint: CheckpointOpt = None,
            pretrained: PretrainedOpt = False,
            num_classes: NumClassesOpt = None,
            in_chans: InChansOpt = None,
            batch_size: BatchSizeOpt = 1,
            input_size: InputSizeOpt = None,
            dynamic_batch: Annotated[
                bool, typer.Option("--dynamic-batch", help="Mark batch axis as dynamic.")
            ] = False,
            device: DeviceOpt = Device.cpu,
            strict: Annotated[
                bool, typer.Option(help="Enable strict graph capture during torch.export.")
            ] = False,
            normalize: NormalizeOpt = False,
            softmax: SoftmaxOpt = False,
            mean: MeanOpt = None,
            std: StdOpt = None,
            verify: Annotated[
                bool,
                typer.Option(help="Load the saved .pt2 archive after export to validate it."),
            ] = True,
        ) -> None:
            if dynamic_batch and batch_size < 2:
                raise ConfigurationError(
                    "--dynamic-batch requires --batch-size >= 2 for stable symbolic shape capture."
                )

            prep = prepare_export(
                model_name=model_name,
                output=output,
                checkpoint=checkpoint,
                pretrained=pretrained,
                num_classes=num_classes,
                in_chans=in_chans,
                batch_size=batch_size,
                input_size=input_size,
                device=device,
                normalize=normalize,
                softmax=softmax,
                mean=mean,
                std=std,
            )

            dynamic_shapes: tuple[dict[int, torch.export.Dim], ...] | None = None
            if dynamic_batch:
                dynamic_shapes = ({0: torch.export.Dim("batch")},)

            try:
                exported_program = torch.export.export(
                    prep.model,
                    (prep.example_input,),
                    dynamic_shapes=dynamic_shapes,
                    strict=strict,
                )
            except Exception as exc:
                raise ExportError(f"torch.export capture failed: {exc}") from exc

            try:
                torch.export.save(exported_program, str(prep.output_path))
            except Exception as exc:
                raise ExportError(f"Failed to save torch.export archive: {exc}") from exc

            if verify:
                try:
                    torch.export.load(str(prep.output_path))
                except Exception as exc:
                    raise ExportError(f"Saved torch.export archive failed to load: {exc}") from exc

        return command



from __future__ import annotations

from collections.abc import Callable
from enum import StrEnum
from pathlib import Path
from typing import Annotated

import torch
import typer

from timmx.errors import ConfigurationError, ExportError
from timmx.export.base import DependencyStatus, ExportBackend
from timmx.export.calibration import resolve_calibration_batches
from timmx.export.common import (
    BatchSizeOpt,
    CheckpointOpt,
    DeviceOpt,
    InChansOpt,
    InputSizeOpt,
    MeanOpt,
    ModelNameArg,
    NormalizeOpt,
    NumClassesOpt,
    OutputOpt,
    PretrainedOpt,
    SoftmaxOpt,
    StdOpt,
    prepare_export,
)
from timmx.export.types import Device


class LiteRTMode(StrEnum):
    fp32 = "fp32"
    fp16 = "fp16"
    dynamic_int8 = "dynamic-int8"
    int8 = "int8"


class LiteRTBackend(ExportBackend):
    name = "litert"
    help = "Export a timm model to LiteRT/TFLite using litert-torch."

    def check_dependencies(self) -> DependencyStatus:
        missing = []
        try:
            import litert_torch  # noqa: F401
        except ImportError:
            missing.append("litert-torch")
        return DependencyStatus(
            available=not missing,
            missing_packages=missing,
            install_hint="pip install 'timmx[litert]'",
        )

    def create_command(self) -> Callable[..., None]:
        def command(
            model_name: ModelNameArg,
            output: OutputOpt,
            checkpoint: CheckpointOpt = None,
            pretrained: PretrainedOpt = False,
            num_classes: NumClassesOpt = None,
            in_chans: InChansOpt = None,
            batch_size: BatchSizeOpt = 1,
            input_size: InputSizeOpt = None,
            device: DeviceOpt = Device.cpu,
            mode: Annotated[
                LiteRTMode, typer.Option(help="Export precision / quantization mode.")
            ] = LiteRTMode.fp32,
            calibration_data: Annotated[
                Path | None,
                typer.Option(
                    help=(
                        "Path to calibration data: an image directory or a "
                        "torch-saved tensor (N, C, H, W). Required for int8 modes "
                        "unless --random-calibration is set."
                    )
                ),
            ] = None,
            calibration_steps: Annotated[
                int | None,
                typer.Option(
                    help=(
                        "Number of calibration batches to consume. "
                        "Default is all full batches from --calibration-data."
                    )
                ),
            ] = None,
            calibration_samples: Annotated[
                int | None,
                typer.Option(
                    help=(
                        "Max number of images to load when --calibration-data is a "
                        "directory. Default is 128."
                    )
                ),
            ] = None,
            random_calibration: Annotated[
                bool,
                typer.Option(
                    "--random-calibration",
                    help=(
                        "Use random noise for calibration instead of real data. "
                        "Not recommended for production use."
                    ),
                ),
            ] = False,
            nhwc_input: Annotated[
                bool,
                typer.Option(help="Expose the first model input as NHWC instead of NCHW."),
            ] = False,
            verify: Annotated[
                bool, typer.Option(help="Load and allocate the exported TFLite model.")
            ] = True,
            normalize: NormalizeOpt = False,
            softmax: SoftmaxOpt = False,
            mean: MeanOpt = None,
            std: StdOpt = None,
        ) -> None:
            int8_modes = {LiteRTMode.dynamic_int8, LiteRTMode.int8}

            if mode in int8_modes and device != Device.cpu:
                raise ConfigurationError(
                    "LiteRT int8 modes currently require --device cpu for PT2E quantization."
                )
            if mode not in int8_modes and (mean is not None or std is not None) and not normalize:
                raise ConfigurationError(
                    "--mean/--std require --normalize unless used for --mode dynamic-int8 "
                    "or --mode int8 calibration."
                )
            if mode not in int8_modes and (
                calibration_data is not None
                or calibration_steps is not None
                or calibration_samples is not None
                or random_calibration
            ):
                raise ConfigurationError(
                    "--calibration-data, --calibration-steps, --calibration-samples, "
                    "and --random-calibration are only valid with "
                    "--mode dynamic-int8 or --mode int8."
                )

            litert_torch = _import_litert_torch()

            prep = prepare_export(
                model_name=model_name,
                output=output,
                checkpoint=checkpoint,
                pretrained=pretrained,
                num_classes=num_classes,
                in_chans=in_chans,
                batch_size=batch_size,
                input_size=input_size,
                device=device,
                normalize=normalize,
                softmax=softmax,
                mean=mean if normalize else None,
                std=std if normalize else None,
            )

            convert_module: torch.nn.Module = prep.model
            quant_config = None
            converter_flags: dict[str, object] = {}
            example_input = prep.example_input

            if mode == LiteRTMode.fp16:
                tensorflow = _import_tensorflow()
                converter_flags = {
                    "optimizations": [tensorflow.lite.Optimize.DEFAULT],
                    "target_spec": {"supported_types": [tensorflow.float16]},
                }
            elif mode in int8_modes:
                calibration_batches = resolve_calibration_batches(
                    calibration_data=calibration_data,
                    calibration_steps=calibration_steps,
                    batch_size=batch_size,
                    input_size=prep.resolved_input_size,
                    device=prep.torch_device,
                    model=prep.model,
                    calibration_samples=calibration_samples,
                    random_calibration=random_calibration,
                    mean=mean,
                    std=std,
                    normalize_images=not normalize,
                )
                example_input = calibration_batches[0]
                convert_module, quant_config = _prepare_pt2e_quantized_module(
                    prep.model,
                    example_input,
                    calibration_batches=calibration_batches,
                    is_dynamic=(mode == LiteRTMode.dynamic_int8),
                )

            if nhwc_input:
                _validate_nhwc_input_compatibility(example_input)
                convert_module = litert_torch.to_channel_last_io(convert_module, args=[0])
                example_input = _to_nhwc_input(example_input)

            try:
                edge_model = litert_torch.convert(
                    convert_module,
                    (example_input,),
                    quant_config=quant_config,
                    _ai_edge_converter_flags=converter_flags,
                )
            except Exception as exc:
                raise ExportError(f"LiteRT conversion failed: {exc}") from exc

            try:
                edge_model.export(str(prep.output_path))
            except Exception as exc:
                raise ExportError(f"Failed to save LiteRT model: {exc}") from exc

            if verify:
                _verify_tflite_model(prep.output_path)

        return command


def _prepare_pt2e_quantized_module(
    model: torch.nn.Module,
    example_input: torch.Tensor,
    *,
    calibration_batches: list[torch.Tensor],
    is_dynamic: bool,
) -> tuple[torch.nn.Module, object]:
    from litert_torch.quantize import pt2e_quantizer, quant_config
    from torchao.quantization.pt2e import quantize_pt2e

    try:
        exported_module = torch.export.export(model, (example_input,), strict=False).module()

        quantizer = pt2e_quantizer.PT2EQuantizer().set_global(
            pt2e_quantizer.get_symmetric_quantization_config(
                is_per_channel=False,
                is_dynamic=is_dynamic,
            )
        )
        prepared_module = quantize_pt2e.prepare_pt2e(exported_module, quantizer)

        with torch.no_grad():
            for calibration_batch in calibration_batches:
                prepared_module(calibration_batch)

        quantized_module = quantize_pt2e.convert_pt2e(prepared_module, fold_quantize=False)
    except Exception as exc:
        mode_label = "dynamic-int8" if is_dynamic else "int8"
        raise ExportError(f"Failed to prepare {mode_label} PT2E quantized model: {exc}") from exc

    return quantized_module, quant_config.QuantConfig(pt2e_quantizer=quantizer)


def _validate_nhwc_input_compatibility(example_input: torch.Tensor) -> None:
    if example_input.ndim < 3:
        raise ConfigurationError(
            "--nhwc-input requires rank >= 3 (for NHWC -> NCHW transposition)."
        )


def _to_nhwc_input(example_input: torch.Tensor) -> torch.Tensor:
    dims = [0, *range(2, example_input.ndim), 1]
    return example_input.permute(*dims).contiguous()


def _verify_tflite_model(output_path: Path) -> None:
    try:
        from ai_edge_litert import interpreter as tfl_interpreter
    except ImportError as exc:
        raise ExportError(
            "ai-edge-litert is required to verify LiteRT export. "
            "Install with: pip install ai-edge-litert"
        ) from exc

    try:
        interpreter = tfl_interpreter.Interpreter(model_path=str(output_path))
        interpreter.allocate_tensors()
    except Exception as exc:
        raise ExportError(f"Saved LiteRT model failed verification: {exc}") from exc


def _import_litert_torch() -> object:
    try:
        import litert_torch
    except ImportError as exc:
        raise ExportError(
            "litert-torch is required for LiteRT export. Install with: pip install 'timmx[litert]'"
        ) from exc
    return litert_torch


def _import_tensorflow() -> object:
    try:
        import tensorflow
    except ImportError as exc:
        raise ExportError(
            "TensorFlow is required for LiteRT fp16 conversion flags. "
            "Install with: pip install tensorflow"
        ) from exc
    return tensorflow



from __future__ import annotations

import tempfile
from collections.abc import Callable
from enum import StrEnum
from pathlib import Path
from typing import Annotated

import torch
import typer

from timmx.errors import ConfigurationError, ExportError
from timmx.export.base import DependencyStatus, ExportBackend
from timmx.export.calibration import resolve_calibration_batches
from timmx.export.common import (
    BatchSizeOpt,
    CheckpointOpt,
    DeviceOpt,
    InChansOpt,
    InputSizeOpt,
    MeanOpt,
    ModelNameArg,
    NormalizeOpt,
    NumClassesOpt,
    OutputOpt,
    PretrainedOpt,
    SoftmaxOpt,
    StdOpt,
    prepare_export,
)
from timmx.export.types import Device

DEFAULT_OPSET = 18
DEFAULT_WORKSPACE_GIB = 4


class TensorRTMode(StrEnum):
    fp32 = "fp32"
    fp16 = "fp16"
    int8 = "int8"


class TensorRTBackend(ExportBackend):
    name = "tensorrt"
    help = "Export a timm model to a TensorRT engine via ONNX."

    def check_dependencies(self) -> DependencyStatus:
        missing = []
        try:
            import tensorrt  # noqa: F401
        except ImportError:
            missing.append("tensorrt")
        try:
            import onnxscript  # noqa: F401
        except ImportError:
            missing.append("onnxscript")
        hints = []
        if "tensorrt" in missing:
            hints.append("pip install tensorrt")
        if "onnxscript" in missing:
            hints.append("pip install 'timmx[onnx]'")
        return DependencyStatus(
            available=not missing,
            missing_packages=missing,
            install_hint=" && ".join(hints) if hints else "",
        )

    def create_command(self) -> Callable[..., None]:
        def command(
            model_name: ModelNameArg,
            output: OutputOpt,
            checkpoint: CheckpointOpt = None,
            pretrained: PretrainedOpt = False,
            num_classes: NumClassesOpt = None,
            in_chans: InChansOpt = None,
            batch_size: BatchSizeOpt = 1,
            input_size: InputSizeOpt = None,
            device: DeviceOpt = Device.cuda,
            mode: Annotated[
                TensorRTMode, typer.Option(help="Engine precision mode.")
            ] = TensorRTMode.fp32,
            workspace: Annotated[
                int, typer.Option(help="Maximum workspace memory in GiB.")
            ] = DEFAULT_WORKSPACE_GIB,
            opset: Annotated[
                int, typer.Option(help="ONNX opset version for intermediate export.")
            ] = DEFAULT_OPSET,
            dynamic_batch: Annotated[
                bool,
                typer.Option(
                    "--dynamic-batch",
                    help="Build engine with dynamic batch size using optimization profiles.",
                ),
            ] = False,
            batch_min: Annotated[
                int, typer.Option(help="Minimum batch size for dynamic batch profile.")
            ] = 1,
            batch_max: Annotated[
                int, typer.Option(help="Maximum batch size for dynamic batch profile.")
            ] = 16,
            calibration_data: Annotated[
                Path | None,
                typer.Option(
                    help=(
                        "Path to calibration data: an image directory or a "
                        "torch-saved tensor (N, C, H, W). Required for --mode int8 "
                        "unless --random-calibration is set."
                    )
                ),
            ] = None,
            calibration_steps: Annotated[
                int | None,
                typer.Option(
                    help=(
                        "Number of calibration batches to consume. "
                        "Default is all full batches from --calibration-data."
                    )
                ),
            ] = None,
            calibration_samples: Annotated[
                int | None,
                typer.Option(
                    help=(
                        "Max number of images to load when --calibration-data is a "
                        "directory. Default is 128."
                    )
                ),
            ] = None,
            random_calibration: Annotated[
                bool,
                typer.Option(
                    "--random-calibration",
                    help=(
                        "Use random noise for calibration instead of real data. "
                        "Not recommended for production use."
                    ),
                ),
            ] = False,
            calibration_cache: Annotated[
                Path | None,
                typer.Option(help="Path to read/write TensorRT INT8 calibration cache."),
            ] = None,
            keep_onnx: Annotated[
                bool,
                typer.Option(
                    "--keep-onnx", help="Keep the intermediate ONNX file alongside the engine."
                ),
            ] = False,
            verbose: Annotated[
                bool, typer.Option(help="Enable verbose TensorRT builder logging.")
            ] = False,
            normalize: NormalizeOpt = False,
            softmax: SoftmaxOpt = False,
            mean: MeanOpt = None,
            std: StdOpt = None,
        ) -> None:
            if device != Device.cuda:
                raise ConfigurationError("TensorRT export requires --device cuda.")

            if opset < 7:
                raise ConfigurationError("--opset must be >= 7.")

            if workspace < 1:
                raise ConfigurationError("--workspace must be >= 1 GiB.")

            if (
                mode != TensorRTMode.int8
                and (mean is not None or std is not None)
                and not normalize
            ):
                raise ConfigurationError(
                    "--mean/--std require --normalize unless used for --mode int8 calibration."
                )

            if mode != TensorRTMode.int8 and (
                calibration_data is not None
                or calibration_steps is not None
                or calibration_samples is not None
                or calibration_cache is not None
                or random_calibration
            ):
                raise ConfigurationError(
                    "--calibration-data, --calibration-steps, --calibration-samples, "
                    "--calibration-cache, and --random-calibration "
                    "are only valid with --mode int8."
                )

            if dynamic_batch:
                if batch_size < 2:
                    raise ConfigurationError(
                        "--batch-size must be >= 2 with --dynamic-batch "
                        "for stable symbolic shape capture."
                    )
                if batch_min < 1:
                    raise ConfigurationError("--batch-min must be >= 1.")
                if batch_max < batch_size:
                    raise ConfigurationError("--batch-max must be >= --batch-size.")
                if batch_min > batch_size:
                    raise ConfigurationError("--batch-min must be <= --batch-size.")

            trt = _import_tensorrt()

            prep = prepare_export(
                model_name=model_name,
                output=output,
                checkpoint=checkpoint,
                pretrained=pretrained,
                num_classes=num_classes,
                in_chans=in_chans,
                batch_size=batch_size,
                input_size=input_size,
                device=device,
                normalize=normalize,
                softmax=softmax,
                mean=mean if normalize else None,
                std=std if normalize else None,
            )

            onnx_path: Path
            temp_dir: tempfile.TemporaryDirectory[str] | None = None

            if keep_onnx:
                onnx_path = prep.output_path.with_suffix(".onnx")
            else:
                temp_dir = tempfile.TemporaryDirectory()
                onnx_path = Path(temp_dir.name) / "model.onnx"

            _require_onnxscript()

            export_kwargs: dict[str, object] = {
                "opset_version": opset,
                "input_names": ["input"],
                "output_names": ["output"],
                "dynamo": True,
                "fallback": True,
                "external_data": False,
            }
            if dynamic_batch:
                batch_dim = torch.export.Dim("batch", min=batch_min, max=batch_max)
                export_kwargs["dynamic_shapes"] = {"x": {0: batch_dim}}

            try:
                torch.onnx.export(
                    prep.model,
                    (prep.example_input,),
                    f=str(onnx_path),
                    **export_kwargs,
                )
            except Exception as exc:
                if temp_dir is not None:
                    temp_dir.cleanup()
                raise ExportError(f"Intermediate ONNX export failed: {exc}") from exc

            try:
                logger = trt.Logger(trt.Logger.VERBOSE if verbose else trt.Logger.WARNING)
                builder = trt.Builder(logger)
                network = builder.create_network(
                    1 << int(trt.NetworkDefinitionCreationFlag.EXPLICIT_BATCH)
                )
                parser = trt.OnnxParser(network, logger)

                with open(onnx_path, "rb") as onnx_file:
                    if not parser.parse(onnx_file.read()):
                        errors = [parser.get_error(i) for i in range(parser.num_errors)]
                        error_msgs = "\n".join(str(e) for e in errors)
                        raise ExportError(f"TensorRT ONNX parsing failed:\n{error_msgs}")

                config = builder.create_builder_config()
                config.set_memory_pool_limit(trt.MemoryPoolType.WORKSPACE, workspace * (1 << 30))

                if mode == TensorRTMode.fp16:
                    config.set_flag(trt.BuilderFlag.FP16)
                elif mode == TensorRTMode.int8:
                    config.set_flag(trt.BuilderFlag.INT8)
                    calibrator = _create_calibrator(
                        trt=trt,
                        calibration_data=calibration_data,
                        calibration_steps=calibration_steps,
                        calibration_cache=calibration_cache,
                        batch_size=batch_size,
                        input_size=prep.resolved_input_size,
                        device=prep.torch_device,
                        model=prep.model,
                        calibration_samples=calibration_samples,
                        random_calibration=random_calibration,
                        mean=mean,
                        std=std,
                        normalize_images=not normalize,
                    )
                    config.int8_calibrator = calibrator

                if dynamic_batch:
                    profile = builder.create_optimization_profile()
                    input_shape_min = (batch_min, *prep.resolved_input_size)
                    input_shape_opt = (batch_size, *prep.resolved_input_size)
                    input_shape_max = (batch_max, *prep.resolved_input_size)
                    profile.set_shape("input", input_shape_min, input_shape_opt, input_shape_max)
                    config.add_optimization_profile(profile)

                serialized_engine = builder.build_serialized_network(network, config)
                if serialized_engine is None:
                    raise ExportError("TensorRT engine build returned None.")

            except ExportError:
                raise
            except Exception as exc:
                raise ExportError(f"TensorRT engine build failed: {exc}") from exc
            finally:
                if temp_dir is not None:
                    temp_dir.cleanup()

            try:
                prep.output_path.write_bytes(serialized_engine)
            except Exception as exc:
                raise ExportError(
                    f"Failed to write TensorRT engine to {prep.output_path}: {exc}"
                ) from exc

        return command


def _make_calibrator_class(trt: object) -> type:
    class _Calibrator(trt.IInt8MinMaxCalibrator):
        def __init__(self, batches: list[torch.Tensor], cache_path: Path) -> None:
            super().__init__()
            self._batches = batches
            self._batch_iter = iter(batches)
            self._batch_size = batches[0].shape[0]
            self._cache_path = cache_path
            self._current_batch: torch.Tensor | None = None

        def get_batch_size(self) -> int:
            return self._batch_size

        def get_batch(self, names: list[str]) -> list[int] | None:
            try:
                batch = next(self._batch_iter)
                self._current_batch = batch.cuda().contiguous()
                return [self._current_batch.data_ptr()]
            except StopIteration:
                return None

        def read_calibration_cache(self) -> bytes | None:
            if self._cache_path.exists():
                return self._cache_path.read_bytes()
            return None

        def write_calibration_cache(self, cache: bytes) -> None:
            self._cache_path.write_bytes(cache)

    return _Calibrator


def _create_calibrator(
    *,
    trt: object,
    calibration_data: Path | None,
    calibration_steps: int | None,
    calibration_cache: Path | None,
    batch_size: int,
    input_size: tuple[int, int, int],
    device: torch.device,
    model: torch.nn.Module,
    calibration_samples: int | None,
    random_calibration: bool,
    mean: tuple[float, ...] | None = None,
    std: tuple[float, ...] | None = None,
    normalize_images: bool = True,
) -> object:
    batches = resolve_calibration_batches(
        calibration_data=calibration_data,
        calibration_steps=calibration_steps,
        batch_size=batch_size,
        input_size=input_size,
        device=device,
        model=model,
        calibration_samples=calibration_samples,
        random_calibration=random_calibration,
        mean=mean,
        std=std,
        normalize_images=normalize_images,
    )
    cache_path = calibration_cache or Path("tensorrt_calibration.cache")
    calibrator_cls = _make_calibrator_class(trt)
    return calibrator_cls(batches=batches, cache_path=cache_path)


def _require_onnxscript() -> None:
    try:
        import onnxscript  # noqa: F401
    except ImportError as exc:
        raise ExportError(
            "onnxscript is required for TensorRT export (dynamo-based ONNX export). "
            "Install with: pip install 'timmx[onnx]'"
        ) from exc


def _import_tensorrt() -> object:
    try:
        import tensorrt as trt
    except ImportError as exc:
        raise ExportError(
            "tensorrt is required for TensorRT export. "
            "Install with: pip install tensorrt (requires CUDA)"
        ) from exc
    return trt



from __future__ import annotations

from abc import ABC, abstractmethod
from collections.abc import Callable
from dataclasses import dataclass


@dataclass
class DependencyStatus:
    """Result of checking whether a backend's dependencies are importable."""

    available: bool
    missing_packages: list[str]
    install_hint: str


class ExportBackend(ABC):
    """Contract implemented by each export backend."""

    name: str
    help: str

    @abstractmethod
    def create_command(self) -> Callable[..., None]:
        """Return a Typer-compatible command function."""

    def check_dependencies(self) -> DependencyStatus:
        """Check if this backend's required dependencies are importable.

        Backends with optional dependencies should override this method.
        The default returns available=True (no extra deps beyond torch).
        """
        return DependencyStatus(available=True, missing_packages=[], install_hint="")



from __future__ import annotations

from collections.abc import Callable
from enum import StrEnum
from typing import Annotated

import typer

from timmx.console import console
from timmx.errors import ConfigurationError, ExportError
from timmx.export.base import DependencyStatus, ExportBackend
from timmx.export.common import (
    BatchSizeOpt,
    CheckpointOpt,
    DeviceOpt,
    InChansOpt,
    InputSizeOpt,
    MeanOpt,
    ModelNameArg,
    NormalizeOpt,
    NumClassesOpt,
    OutputOpt,
    PretrainedOpt,
    SoftmaxOpt,
    StdOpt,
    prepare_export,
)
from timmx.export.types import Device


class ExportSource(StrEnum):
    trace = "trace"
    torch_export = "torch-export"


class ConvertTo(StrEnum):
    mlprogram = "mlprogram"
    neuralnetwork = "neuralnetwork"


class ComputePrecision(StrEnum):
    float16 = "float16"
    float32 = "float32"


class CoreMLBackend(ExportBackend):
    name = "coreml"
    help = "Export a timm model to Core ML."

    def check_dependencies(self) -> DependencyStatus:
        missing = []
        try:
            import coremltools  # noqa: F401
        except ImportError:
            missing.append("coremltools")
        return DependencyStatus(
            available=not missing,
            missing_packages=missing,
            install_hint="pip install 'timmx[coreml]'",
        )

    def create_command(self) -> Callable[..., None]:
        def command(
            model_name: ModelNameArg,
            output: OutputOpt,
            checkpoint: CheckpointOpt = None,
            pretrained: PretrainedOpt = False,
            num_classes: NumClassesOpt = None,
            in_chans: InChansOpt = None,
            batch_size: BatchSizeOpt = 1,
            input_size: InputSizeOpt = None,
            dynamic_batch: Annotated[
                bool,
                typer.Option(
                    "--dynamic-batch", help="Export with flexible Core ML batch dimension."
                ),
            ] = False,
            batch_upper_bound: Annotated[
                int,
                typer.Option(
                    help="Upper bound used for flexible batch when --dynamic-batch is enabled."
                ),
            ] = 8,
            device: DeviceOpt = Device.cpu,
            source: Annotated[
                ExportSource,
                typer.Option(help="Model capture method: trace (default) or torch-export (beta)."),
            ] = ExportSource.trace,
            convert_to: Annotated[
                ConvertTo, typer.Option(help="Core ML model type to generate.")
            ] = ConvertTo.mlprogram,
            compute_precision: Annotated[
                ComputePrecision | None,
                typer.Option(help="Precision for mlprogram conversion."),
            ] = None,
            half: Annotated[
                bool,
                typer.Option("--half", help="Quantize weights to float16."),
            ] = False,
            int8: Annotated[
                bool,
                typer.Option("--int8", help="Quantize weights to 8-bit."),
            ] = False,
            int4: Annotated[
                bool,
                typer.Option("--int4", help="Quantize weights to 4-bit (mlpackage only)."),
            ] = False,
            normalize: NormalizeOpt = False,
            softmax: SoftmaxOpt = False,
            mean: MeanOpt = None,
            std: StdOpt = None,
            verify: Annotated[
                bool, typer.Option(help="Reload the saved Core ML model metadata after export.")
            ] = True,
        ) -> None:
            if convert_to == ConvertTo.neuralnetwork and compute_precision is not None:
                raise ConfigurationError(
                    "--compute-precision is only supported when --convert-to mlprogram."
                )
            if dynamic_batch:
                if batch_upper_bound < 1:
                    raise ConfigurationError("--batch-upper-bound must be >= 1.")
                if batch_upper_bound < batch_size:
                    raise ConfigurationError("--batch-upper-bound must be >= --batch-size.")
            if source == ExportSource.torch_export and dynamic_batch and batch_size < 2:
                raise ConfigurationError(
                    "--dynamic-batch with --source torch-export requires "
                    "--batch-size >= 2 for stable symbolic shape capture."
                )
            if sum([half, int8, int4]) > 1:
                raise ConfigurationError(
                    "Only one of --half, --int8, --int4 can be specified at a time."
                )
            if int4 and convert_to != ConvertTo.mlprogram:
                raise ConfigurationError("--int4 is only supported with --convert-to mlprogram.")

            ct = _import_coremltools()

            if source == ExportSource.torch_export:
                console.print(
                    "[bold yellow]note:[/bold yellow] torch.export support in coremltools "
                    "is beta. If conversion fails, try the default --source trace.",
                    highlight=False,
                )

            prep = prepare_export(
                model_name=model_name,
                output=output,
                checkpoint=checkpoint,
                pretrained=pretrained,
                num_classes=num_classes,
                in_chans=in_chans,
                batch_size=batch_size,
                input_size=input_size,
                device=device,
                normalize=normalize,
                softmax=softmax,
                mean=mean,
                std=std,
            )

            import torch

            if source == ExportSource.torch_export:
                dynamic_shapes: tuple[dict[int, torch.export.Dim], ...] | None = None
                if dynamic_batch:
                    dynamic_shapes = ({0: torch.export.Dim("batch", min=1, max=batch_upper_bound)},)

                try:
                    exported_program = torch.export.export(
                        prep.model,
                        (prep.example_input,),
                        dynamic_shapes=dynamic_shapes,
                    )
                    # coremltools requires ATEN dialect, not TRAINING
                    exported_program = exported_program.run_decompositions({})
                except Exception as exc:
                    raise ExportError(f"torch.export capture failed: {exc}") from exc

                convert_kwargs: dict[str, object] = {
                    "convert_to": str(convert_to),
                }
                if compute_precision is not None:
                    convert_kwargs["compute_precision"] = _map_compute_precision(
                        str(compute_precision), ct
                    )

                try:
                    coreml_model = ct.convert(exported_program, **convert_kwargs)
                except Exception as exc:
                    raise ExportError(f"Core ML conversion failed: {exc}") from exc
            else:
                with torch.no_grad():
                    try:
                        traced_model = torch.jit.trace(prep.model, prep.example_input)
                    except Exception as exc:
                        raise ExportError(f"TorchScript trace failed: {exc}") from exc

                # TODO: Use ImageType with native normalization
                input_type = ct.TensorType(
                    name="input",
                    shape=_build_input_shape(
                        batch_size=batch_size,
                        dynamic_batch=dynamic_batch,
                        batch_upper_bound=batch_upper_bound,
                        input_size=prep.resolved_input_size,
                        ct=ct,
                    ),
                )
                convert_kwargs: dict[str, object] = {
                    "source": "pytorch",
                    "inputs": [input_type],
                    "convert_to": str(convert_to),
                }
                if compute_precision is not None:
                    convert_kwargs["compute_precision"] = _map_compute_precision(
                        str(compute_precision), ct
                    )

                try:
                    coreml_model = ct.convert(traced_model, **convert_kwargs)
                except Exception as exc:
                    raise ExportError(f"Core ML conversion failed: {exc}") from exc

            bits = 4 if int4 else 8 if int8 else 16 if half else 32
            if bits < 32:
                is_mlprogram = convert_to == ConvertTo.mlprogram
                try:
                    coreml_model = _quantize_weights(
                        coreml_model, bits=bits, is_mlprogram=is_mlprogram, ct=ct
                    )
                except Exception as exc:
                    raise ExportError(f"Weight quantization failed: {exc}") from exc

            try:
                coreml_model.save(str(prep.output_path))
            except Exception as exc:
                raise ExportError(f"Failed to save Core ML model: {exc}") from exc

            if verify:
                try:
                    ct.models.MLModel(str(prep.output_path), skip_model_load=True)
                except Exception as exc:
                    raise ExportError(f"Saved Core ML model failed verification: {exc}") from exc

        return command


def _build_input_shape(
    *,
    batch_size: int,
    dynamic_batch: bool,
    batch_upper_bound: int,
    input_size: tuple[int, int, int],
    ct: object,
) -> tuple[object, int, int, int] | tuple[int, int, int, int]:
    if not dynamic_batch:
        return (batch_size, *input_size)

    batch_dim = ct.RangeDim(
        lower_bound=1,
        upper_bound=batch_upper_bound,
        default=batch_size,
        symbol="batch",
    )
    return (batch_dim, *input_size)


def _map_compute_precision(value: str, ct: object) -> object:
    if value == "float16":
        return ct.precision.FLOAT16
    return ct.precision.FLOAT32


def _quantize_weights(coreml_model: object, *, bits: int, is_mlprogram: bool, ct: object) -> object:
    if not is_mlprogram:
        # neuralnetwork: linear quantization (no scikit-learn needed)
        mode = "linear" if bits == 16 else "linear_symmetric"
        console.print(f"[bold]Quantizing weights to {bits}-bit ({mode})...[/bold]")
        return ct.models.neural_network.quantization_utils.quantize_weights(
            coreml_model, bits, mode
        )

    if bits == 16:
        console.print(
            "[bold yellow]note:[/bold yellow] mlprogram models already use float16 weights"
            " — skipping --half.",
            highlight=False,
        )
        return coreml_model

    # mlprogram: k-means palettization for int8/int4
    import coremltools.optimize.coreml as cto

    op_config = cto.OpPalettizerConfig(mode="kmeans", nbits=bits, weight_threshold=512)
    config = cto.OptimizationConfig(global_config=op_config)
    console.print(f"[bold]Palettizing weights to {bits}-bit (k-means)...[/bold]")
    return cto.palettize_weights(coreml_model, config)


def _import_coremltools() -> object:
    try:
        import coremltools as ct
    except ImportError as exc:
        raise ExportError(
            "coremltools is required for Core ML export. Install with: pip install 'timmx[coreml]'"
        ) from exc
    return ct
